@@ -51,6 +51,30 @@
  *   - struct regnames (sname) now has generic pointer to allow
  *     relationships between devs. name is the full name, now.
  * 
+ *  * 2000/07/15: Kurt Garloff <garloff@suse.de>
+ *    - Open device nodes (but sg) with O_NONBLOCK; unfortunately this is not
+ *      honoured by most of Linux' SCSI high-level drivers :-(
+ *    - Add support for OnStream tapes (osst)
+ *  -> Version 2.20
+ * 
+ *  * 2000/09/04: Kurt Garloff <garloff@suse.de>
+ *    - Bugfix for long hostnames from Doug Gilbert (=> 2.21)
+ *    - Fix parsing of scsi.alias file: Broke on missing LF at the end
+ *  -> 2.22
+ * 
+ *  * 2002/07/26: Kurt Garloff <garloff@suse.de>
+ *    - optional (-e) cbtu naming scheme
+ *    - bugfix WRT alias handling of subdevices (parititons, non-rew. tap)
+ *    - Support for large SCSI IDs and LUNs. (from MGE)
+ *    - Support WWID (INQ EVPD 0x83) report + aliases
+ *    - Fix hex number parser
+ *  -> 2.23
+ * 
+ *  * 2002/07/30: Kurt Garloff <garloff@suse.de>
+ *    - Support for /proc/scsi/scsi extensions
+ *    - More sane way of storing permissions
+ *  -> 2.25
+ * 
  */
 
 #include <stdio.h>
@@ -65,13 +89,15 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/ioctl.h>
+#include <netinet/in.h>
+#include <scsi/scsi_ioctl.h>
 
 
-static char rcsid[] ="$Id: scsidev.c,v 1.24 2000/01/28 00:15:30 garloff Exp $";
+static char rcsid[] ="$Id: scsidev.c,v 1.28.2.7 2002/07/27 00:16:10 garloff Exp $";
 static char *versid = "scsidev " VERSION " 2000/01/17";
 static char *copyright = "Copyright: GNU GPL  (see file COPYING)\n" \
 " (w)  1994--1997 Eric Youngdale <eric@andante.org>\n"\
-"      2000        Kurt Garloff   <garoff@suse.de>";
+"      2000--2002 Kurt Garloff   <garloff@suse.de>";
 
 
 #include "config.h"
@@ -97,16 +123,18 @@ int symlink_alias = 0;
 int filemode = 0600;
 int verbose = 0;
 int quiet = 0;
-int maxmiss = 0;
+int maxmiss = 8;
 int force = 0;
 int san_del = 0;
 int no_san = 0;
+int nm_cbtu = 0;
 int supp_rmvbl = 0;
 char * no_serial = "No serial number";
+unsigned long long no_wwid = 0;
 
 #define DEVSCSI "/dev/scsi"
 #define TESTDEV DEVSCSI "/testdev"
-
+#define PROCSCSI "/proc/scsi/scsi"
 
 /*
  * This program builds entries in /dev/scsi that have names that are tied
@@ -131,17 +159,18 @@ typedef struct regnames
     char * model;
     char * rev;
     char * serial;
+    unsigned long long wwid;
     char inq_devtp;
     char rmvbl;
-    char hostnum;
     char unsafe;
     int  hostid;
     int  major;
     int  minor;
     char * hostname;
-    char chan;
-    char id;
-    char lun;
+    int  hostnum;
+    int  chan;
+    int  id;
+    int  lun;
     char partition;
     struct regnames * alias; // TBR
     struct regnames * related;
@@ -156,6 +185,17 @@ sname * reglist = NULL;
 void build_special();
 int inquiry (int, sname *);
 
+#ifndef OSST_MAJOR
+# define OSST_MAJOR 206
+#endif
+
+#define OSST_SUPPORTS(spnt) (! ( memcmp (spnt->manufacturer, "OnStream", 8) || \
+			       ( memcmp (spnt->model, "SC-", 3) && \
+				 memcmp (spnt->model, "DI-", 3) && \
+				 memcmp (spnt->model, "DP-", 3) && \
+				 memcmp (spnt->model, "FW-", 3) && \
+				 memcmp (spnt->model, "USB", 3) ) ) )
+
 
 /*
  * Used to maintain a list of the nodes that we have seen
@@ -167,8 +207,9 @@ void dumplist ()
     sname * pnt;
     for (pnt = reglist; pnt; pnt = pnt->next) 
     {
-	printf ("%s: %s %s %s (%s)\n", pnt->name,
-		pnt->manufacturer, pnt->model, pnt->rev, pnt->serial);
+	printf ("%s: %s %s %s (%s) %Lx\n", pnt->name,
+		pnt->manufacturer, pnt->model, pnt->rev, pnt->serial,
+		pnt->wwid);
 	printf ("  on %s (%d-%x): c%di%dl%d", pnt->hostname,
 		pnt->hostnum, pnt->hostid, pnt->chan, pnt->id, pnt->lun);
 	if (pnt->partition != -1) printf ("p%d", pnt->partition);
@@ -205,6 +246,7 @@ char sname_cmp (sname *sp1, sname *sp2)
     if (strcmp (sp1->model, sp2->model)) return 7;
     if (strcmp (sp1->rev, sp2->rev)) return 8;
     if (strcmp (sp1->serial, sp2->serial)) return 9;
+    if (sp1->wwid != sp2->wwid) return 10;
     return 0;
 }
 
@@ -236,6 +278,7 @@ sname * register_dev (char * name, int major, int minor,
      * Initialize this - they may be needed later.
      */
     spnt->model = spnt->manufacturer = spnt->serial = spnt->rev = NULL;
+    spnt->wwid = no_wwid;
     spnt->next = reglist; reglist = spnt;
     return spnt;
 }
@@ -348,6 +391,10 @@ char * scsiname (sname *spnt)
       case SCSI_TAPE_MAJOR:
 	if (spnt->minor & 0x80) strcat (nm, "n");
 	devchar = 't'; break;
+      case OSST_MAJOR:
+	if (spnt->minor & 0x80) strcat (nm, "nos");
+	else strcat (nm, "os");
+	devchar = 't'; break;
       default:
 	if (!(SCSI_DISK_MAJOR (spnt->major)))
 	{
@@ -360,9 +407,14 @@ char * scsiname (sname *spnt)
 	    sprintf (app, "p%d", spnt->minor % 0x10);
     }
     genpart = nm + strlen (nm);
-    sprintf (genpart, "s%ch%d-%xc%di%dl%d", devchar,
-	     spnt->hostnum, spnt->hostid,
-	     spnt->chan, spnt->id, spnt->lun);
+    if (nm_cbtu) 
+	sprintf (genpart, "s%cc%db%dt%du%d", devchar,
+		 spnt->hostnum, 
+		 spnt->chan, spnt->id, spnt->lun);
+    else
+	sprintf (genpart, "s%ch%d-%xc%di%dl%d", devchar,
+		 spnt->hostnum, spnt->hostid,
+		 spnt->chan, spnt->id, spnt->lun);
     if (*app) strcat (genpart, app);
     spnt->name = strdup (nm);
     return spnt->name;
@@ -386,6 +438,12 @@ char * oldscsiname (sname *spnt)
 	else
 	    sprintf (genpart, "st%d", spnt->minor & 0x7f);
 	break;
+      case OSST_MAJOR:
+	if (spnt->minor & 0x80) 
+	    sprintf (genpart, "nosst%d", spnt->minor & 0x7f);
+	else
+	    sprintf (genpart, "osst%d", spnt->minor & 0x7f);
+	break;
       default:
 	if (!(SCSI_DISK_MAJOR (spnt->major)))
 	{
@@ -394,7 +452,7 @@ char * oldscsiname (sname *spnt)
 	    abort ();
 	}
 		
-	/* SCSI_DISK1 -- SCSI_DISK7 are contiguousin Linux-2.2+ */
+	/* SCSI_DISK1 -- SCSI_DISK7 are contiguous in Linux-2.2+ */
 	if (spnt->major == SCSI_DISK0_MAJOR)
 	    diskno = spnt->minor >> 4;
 	else
@@ -557,6 +615,7 @@ int getscsiinfo (int fd, sname *spnt)
 	
     *(int*)hostname = 63;
     status = ioctl (fd, SCSI_IOCTL_PROBE_HOST, hostname);
+    hostname[63] = '\0';
     //close (fd);
     spnt->hostname = strdup (hostname);
 	
@@ -572,12 +631,12 @@ int getscsiinfo (int fd, sname *spnt)
     if (verbose == 2)
 	fprintf (stderr, "Found %x with idlun %x\n", 
 		 makedev (spnt->major, spnt->minor), id[0]);
-    spnt->hostnum = id[0] >> 24; 
-    spnt->chan    = id[0] >> 16; //()&0xff
-    spnt->lun     = id[0] >>  8;
-    spnt->id      = id[0];
+    spnt->hostnum = id[0] >> 24 & 0xff;
+    spnt->chan    = id[0] >> 16 & 0xff;
+    spnt->lun     = id[0] >>  8 & 0xff;
+    spnt->id      = id[0]       & 0xff;
     spnt->hostid  = id[1];
-	
+
     status = inquiry (fd, spnt);
     scsiname (spnt);
     return status;
@@ -602,12 +661,12 @@ int build_disk (sname * spnt, int no)
     create_dev (spnt1);
     spnt->related = spnt1;
     /* Check if device is there (i.e. medium inside) */
-    fd =  open (spnt1->name, O_RDONLY);
+    fd =  open (spnt1->name, O_RDONLY | O_NONBLOCK);
     /* No access to medium / part. table */
     if (fd < 0) {
 	spnt1->unsafe = 1;
 	/* If it is a removable device, we can't do much more than 
-	 * trusting it or not support it at all*/
+	 * trusting it or not support it at all */
 	if (spnt1->rmvbl && supp_rmvbl) return 0;
 	fprintf (stderr, "Can't access %sremovable %s, which should "
 		 "be equal to %s!\n", (spnt1->rmvbl? "": "NON-"), 
@@ -645,7 +704,7 @@ int build_disk (sname * spnt, int no)
 	    
 	status = mknod ( TESTDEV, 0600 | S_IFBLK,
 			 makedev (spnt1->major, minor) );
-	fd = open (TESTDEV, O_RDONLY);
+	fd = open (TESTDEV, O_RDONLY | O_NONBLOCK);
 	unlink (TESTDEV);
 	if (fd < 0) continue;
 	// TO DO: Add sanity checks here ??
@@ -670,7 +729,7 @@ int build_tape (sname * spnt, int no)
     spnt1->next = reglist; reglist = spnt1;
     create_dev (spnt1);
     /* Check if device is there (i.e. medium inside) */
-    fd =  open (spnt1->name, O_RDONLY);
+    fd =  open (spnt1->name, O_RDONLY | O_NONBLOCK);
     if (fd < 0) {
 	/* Tapes are always accessible, as they are char devices */
 	fprintf (stderr, "Can't access tape %s, which should "
@@ -706,6 +765,56 @@ int build_tape (sname * spnt, int no)
     return 0;
 }
 
+int build_os_tape (sname * spnt, int no)
+{
+    int fd; int status;
+    sname * spnt1 = sname_dup (spnt);
+    spnt1->major = OSST_MAJOR;
+    spnt1->minor = no;
+    scsiname (spnt1);
+    spnt1->next = reglist; reglist = spnt1;
+    create_dev (spnt1);
+    /* Check if device is there (i.e. medium inside) */
+    fd =  open (spnt1->name, O_RDONLY | O_NONBLOCK);
+    if (fd < 0) {
+	/* OnStream tapes are NOT always accessible, as they have a heavy open() function */
+	spnt1->unsafe = 1;
+	if (!quiet || !supp_rmvbl) 
+		    fprintf (stderr, "Can't access tape %s, which should "
+			     "be equal to %s!\n", strrchr (spnt1->name, '/') + 1,
+			     strrchr (spnt->name, '/') + 1);
+	if (supp_rmvbl) goto osst_force_success;
+	/* We don't unlink spnt1->name! Let sanitize take care of it ... */
+	reglist = spnt1->next; free (spnt1->name); 
+	spnt->related = 0; free (spnt1);
+	return 1;
+    }
+    /* Do a sanity check here */
+    status = getscsiinfo (fd, spnt1);
+    close (fd);
+    if (status) fprintf (stderr, "scsidev: Strange: Could not get info from %s\n",
+			 strrchr (spnt1->name, '/') + 1);
+    if (sname_cmp (spnt, spnt1))
+    {
+	fprintf (stderr, "scsidev: What's going on? Dev %s is different from %s\n", 
+		 strrchr (spnt1->name, '/') + 1, strrchr (spnt->name, '/') + 1);
+	spnt -> related = 0; spnt1 -> related = 0;
+	/* And now ? */
+	reglist = spnt1->next; free (spnt1->name); 
+	spnt->related = 0; free (spnt1);
+	return 1;
+    }
+osst_force_success:
+    /* And create the no-rewind alias */
+    spnt1 = sname_dup (spnt1);
+    spnt1->minor |= 0x80;
+    scsiname (spnt1);
+    spnt1->next = reglist; reglist = spnt1;
+    create_dev (spnt1);
+	
+    return 0;
+}
+
 int build_cdrom (sname * spnt, int no)
 {
     int fd; int status;
@@ -715,7 +824,7 @@ int build_cdrom (sname * spnt, int no)
     scsiname (spnt1);
     spnt1->next = reglist; reglist = spnt1;
     create_dev (spnt1);
-    fd =  open (spnt1->name, O_RDONLY);
+    fd =  open (spnt1->name, O_RDONLY | O_NONBLOCK);
     /* No access to medium / part. table */
     if (fd < 0) {
 	spnt1->unsafe = 1;
@@ -797,6 +906,135 @@ void build_sgdevlist ()
 	status = getscsiinfo (fd, spnt);
 	close (fd);
 
+	if (status) { 
+		free (spnt); miss++;
+		if (miss > maxmiss) break;
+		else { minor++; continue; }
+	}
+	scsiname (spnt);
+
+	spnt->next = reglist; reglist = spnt;
+	create_dev (spnt);
+
+	if (!quiet) printf ("Found %s (Type %02x) %c on %s \n", spnt->name,
+			    spnt->inq_devtp, (spnt->rmvbl? 'R' : ' '),
+			    spnt->hostname);
+
+	/* Now register cdroms, tapes, and disks as well */
+	switch (spnt->inq_devtp) {
+	  case TYPE_DISK:
+	  case TYPE_MOD:
+	    if (!build_disk (spnt, disks)) disks++;
+	    else if (!build_disk (spnt, disks+1)) disks += 2;
+	    break;
+	  case TYPE_TAPE:
+	    if (OSST_SUPPORTS(spnt)) 
+	    {
+		if (!build_os_tape (spnt, tapes)) tapes++;
+		else if (!build_os_tape (spnt, tapes+1)) tapes += 2;
+	    }		    
+	    else
+	    {
+		if (!build_tape (spnt, tapes)) tapes++;
+		else if (!build_tape (spnt, tapes+1)) tapes += 2;
+	    }
+	    break;
+	  case TYPE_ROM:
+	  case TYPE_WORM:
+	    if (!build_cdrom (spnt, cdroms)) cdroms++;
+	    else if (!build_cdrom (spnt, cdroms+1)) cdroms += 2;
+	    break;
+	  default:
+	    ;/* nothing to be done */
+	}
+	minor += 1;
+    }
+    //unlink (TESTDEV);
+}
+
+#if 0
+
+/* Test for availability of /proc/scsi/scsi extensions */
+int scsi_ext_status ()
+{
+}
+
+/* Try to switch on extensions */
+void scsi_ext_on ()
+{
+}
+
+/* Try to switch on extensions */
+void scsi_ext_off ()
+{
+}
+
+void build_sgdevlist_procscsi ()
+{
+    int fd; 
+    struct stat statbuf;
+    int status;
+    sname * spnt;
+    int miss = 0;	
+    int minor = 0; int major = SCSI_GENERIC_MAJOR; 
+    char devchar = 'g'; int mode = O_RDWR;
+    int devtype = (SCSI_BLK_MAJOR(major)? S_IFBLK: S_IFCHR);
+    char* buf; int buflen;
+    FILE* mapfile;
+
+    status = stat (DEVSCSI, &statbuf);
+    if (status == -1)
+	return;
+
+    status = stat (TESTDEV, &statbuf);
+    if (status == 0)
+	unlink (TESTDEV);
+
+    if (verbose >= 1)
+	fprintf (stderr, "Building list for s%c (%s dev major %i)\n",
+		 devchar, (SCSI_BLK_MAJOR(major)? "block": "char"), major);
+    
+    mapfile = fopen (PROCSCSI, "r");
+    if (!mapfile) {
+	fprintf (stderr, "scsidev: could not open " PROCSCSI ": %s\n",
+		 strerror (errno));
+	return;
+    }
+    
+    buf = malloc(128); buflen = 128;
+
+	
+	
+    while (getline (&buf, &buflen, mapfile) != -1)
+    {
+	int h,c,i,l;
+	int inqtp, onl;
+	char *nm, *dev, *rest;
+	if (buf[0] == '#')
+		continue;
+	
+	errno = 0;
+	status = mknod ( TESTDEV, 0600 | devtype ,
+			 makedev (major, minor) );
+	if (status) { perror ("scsidev: mknod"); exit (3); }
+	fd = open (TESTDEV, mode);
+	unlink (TESTDEV);
+	if (fd == -1) {
+	    if (verbose == 2)
+	    {
+		fprintf (stderr, "open(%x/%x) returned %d (%d)\n",
+			 major, minor, fd, errno);
+	    }
+	    miss++;
+	    if (miss > maxmiss) break;
+	    else { minor++; continue; }
+	}
+	spnt = (sname*) malloc (sizeof (sname));
+	spnt->major = major;   spnt->minor = minor;
+	spnt->name  = TESTDEV; spnt->partition = -1;
+	status = getscsiinfo (fd, spnt);
+	close (fd);
+
 	if (status) { free (spnt); return; };
 	scsiname (spnt);
 
@@ -814,9 +1052,17 @@ void build_sgdevlist ()
 	    if (!build_disk (spnt, disks)) disks++;
 	    else if (!build_disk (spnt, disks+1)) disks += 2;
 	    break;
-	  case TYPE_TAPE: 
-	    if (!build_tape (spnt, tapes)) tapes++;
-	    else if (!build_tape (spnt, tapes+1)) tapes += 2;
+	  case TYPE_TAPE:
+	    if (OSST_SUPPORTS(spnt)) 
+	    {
+		if (!build_os_tape (spnt, tapes)) tapes++;
+		else if (!build_os_tape (spnt, tapes+1)) tapes += 2;
+	    }		    
+	    else
+	    {
+		if (!build_tape (spnt, tapes)) tapes++;
+		else if (!build_tape (spnt, tapes+1)) tapes += 2;
+	    }
 	    break;
 	  case TYPE_ROM:
 	  case TYPE_WORM:
@@ -829,8 +1075,24 @@ void build_sgdevlist ()
 	minor += 1;
     }
     //unlink (TESTDEV);
+    free(buf);
 }
 
+int try_procscsi ()
+{
+	int se_status = scsi_ext_status();
+	if (!se_status)
+		scsi_ext_on();
+	if (!scsi_ext_status())
+		return -1;
+	
+	build_sgdevlist_procscsi();
+	
+	if (!se_status)
+		scsi_ext_off();
+}
+#endif
+
 
 void usage()
 {
@@ -841,9 +1103,10 @@ void usage()
     fprintf (stderr, " -d     : sanitize by Deleting undetected entries (def: minor->255)\n");
     fprintf (stderr, " -l/-L  : create symLinks for device names / alias names\n");
     fprintf (stderr, " -m mode: permissions to create dev nodes with\n");
-    fprintf (stderr, " -s     : list Serial numbers of devices (if available)\n");
+    fprintf (stderr, " -s     : list Serial numbers /WWIDs of devices (if available)\n");
     fprintf (stderr, " -c mxms: Continue scanning until mxms missing devs found\n");
     fprintf (stderr, " -r     : trust Removeable media (only safe after boot)\n");
+    fprintf (stderr, " -e     : use dEvfs like naming  (cbtu chars)\n");
     fprintf (stderr, " -v/-q  : Verbose/Quiet operation\n");
     fprintf (stderr, " -h     : print Help and exit.\n");
 }
@@ -866,7 +1129,7 @@ int main (int argc, char * argv[])
 	fprintf(stderr, DEVSCSI " either does not exist, or is not a directory\n");
 	exit(0);
     }
-    while ((c = getopt(argc, argv, "flLvqshndrm:c:")) != EOF)
+    while ((c = getopt(argc, argv, "flLvqshnderm:c:")) != EOF)
     {
 	switch (c)
 	{
@@ -888,6 +1151,8 @@ int main (int argc, char * argv[])
 	    san_del = 1; break;
 	  case 'r':
 	    supp_rmvbl = 1; break;
+	  case 'e':
+	    nm_cbtu = 1; break;
 	  case 'v':
 	    verbose++; break;
 	  case 'h':
@@ -904,22 +1169,24 @@ int main (int argc, char * argv[])
     if( force ) flush_sdev ();
 
 #ifdef DEBUG
-    register_dev("/dev/scsi/sdh4-334c0i0l0",   8,  0, 6, 0x334, 0, 0, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/sdh4-334c0i0l0p1", 8,  1, 6, 0x334, 0, 0, 0,  1, "debug", NULL);
-    register_dev("/dev/scsi/sdh4-334c0i0l0p2",8,  2,6, 0x334, 0, 0, 0,  2, "debug", NULL);
-    register_dev("/dev/scsi/sdh4-334c0i0l0p3",8,  3,6, 0x334, 0, 0, 0,  3, "debug", NULL);
-    register_dev("/dev/scsi/sgh4-334c0i0l0", 21,  0,6, 0x334, 0, 0, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/sgh4-334c0i2l0", 21,  1,6, 0x334, 0, 2, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/sgh4-334c0i5l0", 21,  2,6, 0x334, 0, 5, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/srh4-334c0i2l0", 11,  0,6, 0x334, 0, 2, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/sth4-334c0i5l0",  9,  0,6, 0x334, 0, 5, 0, -1, "debug", NULL);
-    register_dev("/dev/scsi/rsth4-334c0i5l0", 9,128,6, 0x334, 0, 5, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/sdh4-334c0i0l0",  8,  0, 6, 0x334, 0, 0, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/sdh4-334c0i0l0p1",8,  1, 6, 0x334, 0, 0, 0,  1, "debug", NULL);
+    register_dev("/dev/scsi/sdh4-334c0i0l0p2",8,  2, 6, 0x334, 0, 0, 0,  2, "debug", NULL);
+    register_dev("/dev/scsi/sdh4-334c0i0l0p3",8,  3, 6, 0x334, 0, 0, 0,  3, "debug", NULL);
+    register_dev("/dev/scsi/sgh4-334c0i0l0", 21,  0, 6, 0x334, 0, 0, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/sgh4-334c0i2l0", 21,  1, 6, 0x334, 0, 2, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/sgh4-334c0i5l0", 21,  2, 6, 0x334, 0, 5, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/srh4-334c0i2l0", 11,  0, 6, 0x334, 0, 2, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/sth4-334c0i5l0",  9,  0, 6, 0x334, 0, 5, 0, -1, "debug", NULL);
+    register_dev("/dev/scsi/rsth4-334c0i5l0", 9,128, 6, 0x334, 0, 5, 0, -1, "debug", NULL);
 #else
     build_sgdevlist ();
 #endif
 
     if( show_serial )
     {
+	if (verbose)
+	    dumplist();
 	for (spnt = reglist; spnt; spnt = spnt->next)
 	{
 	    if( spnt->partition != -1 ) continue;
@@ -931,6 +1198,8 @@ int main (int argc, char * argv[])
 		printf("Device  %s has no serial number\n", spnt->name);
 	    else
 		printf("Serial number of %s: \"%s\"\n", spnt->name, spnt->serial);
+	    if ( spnt->wwid != no_wwid )
+		printf (" WWID: %Lx\n", spnt->wwid);
 	}
     }
 
@@ -976,6 +1245,58 @@ char * get_string (char * pnt, char **result)
     return pnt;
 }
 
+char * get_llnumber (char * pnt, unsigned long long * result)
+{
+    int base = 10;
+    unsigned long long num;
+
+    while (*pnt == ' ' || *pnt == '\t') pnt++;
+    if (pnt[0] == '0' && pnt[1] == 'x') { base = 16; pnt += 2; }
+
+    num = 0;
+    while (1)
+    {
+	if (base == 10 && *pnt >= '0' && *pnt <= '9' )
+	{
+	    num = num * 10ULL + *pnt - '0';
+	    pnt++;
+	    continue;
+	}
+	if ( base == 16 )
+	{
+	    if (*pnt >= '0' && *pnt <= '9')
+	    {
+		num = (num << 4) + *pnt - '0';
+		pnt++;
+		continue;
+	    }
+	    if (*pnt >= 'a' && *pnt <= 'f')
+	    {
+		num = (num << 4) + *pnt - 'a' + 10;
+		pnt++;
+		continue;
+	    }
+	    if (*pnt >= 'A' && *pnt <= 'F')
+	    {
+		num = (num << 4) + *pnt - 'A' + 10;
+		pnt++;
+		continue;
+	    }
+	    break;
+	}
+	/*
+	 * Isn't a digit.  Must be the end of the number.
+	 */
+	break;
+    }
+    while (*pnt == ' ' || *pnt == '\t') pnt++;
+    if (*pnt == ',') pnt++;
+    while (*pnt == ' ' || *pnt == '\t') pnt++;
+    *result = num;
+    return pnt;
+
+}
+
 char * get_number (char * pnt, int * result)
 {
     int base = 10;
@@ -997,27 +1318,27 @@ char * get_number (char * pnt, int * result)
 	{
 	    if (*pnt >= '0' && *pnt <= '9')
 	    {
-		num = num * 10 + *pnt - '0';
+		num = (num << 4) + *pnt - '0';
 		pnt++;
 		continue;
 	    }
 	    if (*pnt >= 'a' && *pnt <= 'f')
 	    {
-		num = num * 10 + *pnt - 'a' + 10;
+		num = (num << 4) + *pnt - 'a' + 10;
 		pnt++;
 		continue;
 	    }
 	    if (*pnt >= 'A' && *pnt <= 'F')
 	    {
-		num = num * 10 + *pnt - 'A' + 10;
+		num = (num << 4) + *pnt - 'A' + 10;
 		pnt++;
 		continue;
 	    }
 	    break;
 	}
-				/*
-				 * Isn't a digit.  Must be the end of the number.
-				 */
+	/*
+	 * Isn't a digit.  Must be the end of the number.
+	 */
 	break;
     }
     while (*pnt == ' ' || *pnt == '\t') pnt++;
@@ -1044,9 +1365,10 @@ char * get_number (char * pnt, int * result)
  * MANUFACTURER="string"
  * MODEL="string"
  * SERIAL_NUMBER="string"	(for those devices that support this).
+ * WWID=number			( "    "      "     "      "      " )
  * REV="string"
  * NAME="string" (alias)
- * DEVTYPE="disk", "tape", "generic", or "cdrom".
+ * DEVTYPE="disk", "tape", "osst", "generic", or "cdrom".
  */
 	
 void build_special ()
@@ -1061,6 +1383,7 @@ void build_special ()
 
     int lun, chan, id, part, hostid, hostnum;
     int line;
+    unsigned long long wwid;	/* host byte order ... */
     int devtype_i;
     char *manufacturer, *model, *serial_number, *name, *devtype, *rev, *host;
     char *scsialias;
@@ -1080,32 +1403,33 @@ void build_special ()
     line = 0;
     while (1)
     {
+	*buffer = 0;
 	fgets (buffer, sizeof(buffer), configfile);
 	line++;
-	if (feof (configfile)) break;
-
-				/*
-				 * Remove trailing \n, if present.
-				 */
+	if (feof (configfile) && !*buffer) break;
+	
+	/*
+	 * Remove trailing \n, if present.
+	 */
 	pnt = buffer + strlen(buffer) - 1;
 	if( *pnt == '\n' ) *pnt = '\0';
 
-				/*
-				 * First, tokenize the input line, and pick out the parameters.
-				 */
+	/*
+	 * First, tokenize the input line, and pick out the parameters.
+	 */
 	lun = -1; id = -1;
 	chan = -1;
 	hostid = -1; hostnum = -1;
-	part = -1;
+	part = -1; wwid = no_wwid;
 	host = NULL;
 	manufacturer = NULL; model = NULL;
 	serial_number = NULL; rev = NULL;
-	name = NULL;
+	name = NULL; 
 	devtype = NULL;
 	pnt = buffer;
 	while (*pnt == ' ' || *pnt == '\t') pnt++;
 
-				/* allow blank lines and comments... */
+	/* allow blank lines and comments... */
 	if( *pnt == '\0' ) continue;
 	if( *pnt == '#' ) continue;
 
@@ -1128,6 +1452,10 @@ void build_special ()
 	    {
 		pnt = get_string(pnt1 + 1, &serial_number);
 	    }
+	    else if ( strcmp(pnt, "wwid") == 0 )
+	    {
+		pnt = get_llnumber(pnt1 + 1, &wwid);
+	    }
 	    else if ( strncmp(pnt, "rev", 3) == 0 )
 	    {
 		pnt = get_string(pnt1 + 1, &rev);
@@ -1196,6 +1524,8 @@ void build_special ()
 	    devtype_i = SCSI_CDROM_MAJOR;
 	else if( strcmp(devtype, "tape") == 0)
 	    devtype_i = SCSI_TAPE_MAJOR;
+	else if( strcmp(devtype, "osst") == 0)
+	    devtype_i = OSST_MAJOR;
 	else if(strcmp(devtype, "generic") == 0 )
 	    devtype_i = SCSI_GENERIC_MAJOR;
 	else
@@ -1205,10 +1535,10 @@ void build_special ()
 	    continue;
 	}
 
-				/*
-				 * OK, minimal requirements are met.  Try and match this to something
-				 * we know about already.
-				 */
+	/*
+	 * OK, minimal requirements are met.  Try and match this to something
+	 * we know about already.
+	 */
 	match = NULL;
 	for (spnt = reglist; spnt; spnt = spnt->next)
 	{
@@ -1226,8 +1556,9 @@ void build_special ()
 		(devtype_i != SCSI_DISK0_MAJOR || ! SCSI_DISK_MAJOR(spnt->major)) )
 		continue;
 	    if( part != spnt->partition ) continue;
-	    if( spnt->major == SCSI_TAPE_MAJOR 
+	    if( (spnt->major == SCSI_TAPE_MAJOR || spnt->major == OSST_MAJOR)
 		&& (spnt->minor & 0x80) != 0) continue;
+	    if( wwid != no_wwid && wwid != spnt->wwid ) continue;
 
 	    /*
 	     * OK, that matches, now obtain some of the strings
@@ -1268,12 +1599,12 @@ void build_special ()
 		match = spnt;
 	}
 
-				/*
-				 * See if there was a non-unique mapping.  If so, then
-				 * don't do anything for this one.
-				 */
-
-				// detect break
+	/*
+	 * See if there was a non-unique mapping.  If so, then
+	 * don't do anything for this one.
+	 */
+	    
+	// detect break
 	if( spnt != NULL )
 	    continue;
 
@@ -1294,7 +1625,7 @@ void build_special ()
 		else fprintf (stderr, "\n");
 	    }
 
-	    if( devtype_i == SCSI_DISK0_MAJOR
+	    if( SCSI_DISK_MAJOR(devtype_i)
 		|| devtype_i == SCSI_CDROM_MAJOR )
 		type = S_IFBLK;
 	    else
@@ -1305,7 +1636,7 @@ void build_special ()
 	     * Just create it.
 	     */
 	    sprintf (scsidev, DEVSCSI "/%s", name);
-	    register_dev (scsidev, match->major, match->minor | 0x80, 
+	    register_dev (scsidev, match->major, match->minor, 
 			  0, 0, 0, 0, 0, 0, 0, match);
 	    if (symlink_alias) {
 		unlink (scsidev);
@@ -1315,7 +1646,7 @@ void build_special ()
 		update_device (scsidev, type, 
 			       match->major, match->minor);
 
-	    if( devtype_i == SCSI_TAPE_MAJOR )
+	    if( devtype_i == SCSI_TAPE_MAJOR || devtype_i == OSST_MAJOR )
 	    {
 		sprintf (scsidev, DEVSCSI "/n%s", name);
 		register_dev (scsidev, match->major, match->minor | 0x80, 
@@ -1353,7 +1684,7 @@ void build_special ()
 
 		    sprintf(scsidev, DEVSCSI "/%s-p%d", name, 
 			    spnt->partition);
-		    register_dev (scsidev, match->major, match->minor | 0x80, 
+		    register_dev (scsidev, match->major, spnt->minor, 
 				  0, 0, 0, 0, 0, 0, 0, spnt);
 		    if (symlink_alias) {
 			unlink (scsidev);
@@ -1398,6 +1729,48 @@ char* getstr (char* page, int start, int stop)
     return str;
 }
 
+unsigned long long extract_wwid (unsigned char* page)
+{
+	unsigned int hi, lo;
+	if (page[1] != 0x83)
+		return no_wwid;
+	/* We only support binary data */
+	if ((page[4] & 0x0f) != 1)
+		return no_wwid;
+	/* We support identifier types 2, 3 */
+	if ((page[5] & 0x0f) != 3 && (page[5] & 0x0f) != 2)
+		return no_wwid;
+	/* Id length should be 8 (64 bits) */
+	if (page[7] != 8)
+		return no_wwid;
+	/* Now translate network byte order to host byte order */
+	hi = ntohl (*(unsigned int*)(page+ 8));
+	lo = ntohl (*(unsigned int*)(page+12));
+	return ((unsigned long long) hi << 32) + (unsigned long long) lo;
+}
+
+#define INQBUFSZ 512
+int get_inq_page (int file, int lun, unsigned char* buf, unsigned char page, char evpd)
+{
+	unsigned char *cmd;
+	
+	memset (buf, 0, INQBUFSZ);
+	
+	*( (int *)  buf)	= 0;			/* length of input data */
+	*( ((int *) buf) + 1 )	= INQBUFSZ-32;	/* length of output buffer */
+	
+	cmd = (char *) ( ((int *) buf) + 2 );
+	
+	cmd[0] = 0x12;			/* INQUIRY */
+	cmd[1] = (lun << 5) | (evpd? 1: 0);	/* lun, evpd=... */
+	cmd[2] = page;			/* page code 0: std inquiry */
+	cmd[3] = 0x00;			/* (reserved) */
+	cmd[4] = 0xfc;			/* allocation length */
+	cmd[5] = 0x00;			/* control */
+	
+	return ioctl (file, SCSI_IOCTL_SEND_COMMAND, buf);
+}
+	
 int inquiry (int infile, sname * spnt)
 {
 #ifdef DEBUG
@@ -1416,73 +1789,68 @@ int inquiry (int infile, sname * spnt)
     }
 #else
     int status;
-    unsigned char *cmd;
-    unsigned char * pagestart;
-    unsigned char buffer[1024];
+    unsigned char buffer[INQBUFSZ];
+    unsigned char * const pagestart = buffer + 8;
     //int infile;
-
-    memset (buffer, 0, sizeof(buffer));
-  
-    *( (int *)  buffer )		= 0;	/* length of input data */
-    *( ((int *) buffer) + 1 )	= sizeof(buffer)-16;	/* length of output buffer */
-
-    cmd = (char *) ( ((int *) buffer) + 2 );
-  
-    cmd[0] = 0x12;			/* INQUIRY */
-    cmd[1] = (spnt->lun << 5) | 0x01;	/* lun, evpd=1 */
-    cmd[2] = 0x80;			/* page code = 0x80, serial number */
-    cmd[3] = 0x00;			/* (reserved) */
-    cmd[4] = 0xff;			/* allocation length */
-    cmd[5] = 0x00;			/* control */
-
+    char have_ser_page = 0;
+    char have_wwid_page = 0;
+    int ln; int off;
+    int lun; int ansi;
+	
     //infile = open(spnt->name, O_RDWR);
     if( infile == -1 ) return -1;
 
-    status = ioctl( infile, 1, buffer );
-    //close(infile);
-    pagestart = buffer + 8;
-
-    if (status) 
-	spnt->serial = no_serial;
-    else
-	spnt->serial = getstr (pagestart, 4, 3+pagestart[3]);
-
-    if (verbose == 2) printf ("Serial for %s: %s\n", spnt->name, spnt->serial);
-
-    memset (buffer, 0, sizeof(buffer));
-  
-    *( (int *)  buffer )		= 0;	/* length of input data */
-    *( ((int *) buffer) + 1 )	= sizeof(buffer)-16;	/* length of output buffer */
-
-    cmd = (char *) ( ((int *) buffer) + 2 );
-  
-    cmd[0] = 0x12;			/* INQUIRY */
-    cmd[1] = (spnt->lun << 5);		/* lun, evpd=0 */
-    cmd[2] = 0x00;			/* page code 0: std inquiry */
-    cmd[3] = 0x00;			/* (reserved) */
-    cmd[4] = 0xff;			/* allocation length */
-    cmd[5] = 0x00;			/* control */
-
-    //infile = open (spnt->name, O_RDWR);
-    //if( infile == -1 ) return -1;
-
-    status = ioctl( infile, 1, buffer );
-    //close(infile);
-    pagestart = buffer + 8;
+    // Std. inquiry
+    status = get_inq_page (infile, 0, buffer, 0, 0);
 
     if (status) 
     { 
-	fprintf (stderr, "INQUIRY failed for %s (%i-%i)!\n", 
-		 spnt->name, spnt->id, spnt->lun);
-	return -1; 
+	fprintf (stderr, "INQUIRY failed for %s (%i-%i/%02x:%02x)!\n",
+		 spnt->name, spnt->id, spnt->lun, spnt->major, spnt->minor);
+	return -1;
     }
-    if (verbose == 2) dumppage (pagestart);
+    if (verbose == 2) 
+		dumppage (pagestart);
     spnt->manufacturer = getstr (pagestart, 8, 15); 
     spnt->model = getstr (pagestart, 16, 31);
     spnt->rev = getstr (pagestart, 32, 35);
     spnt->inq_devtp = pagestart[0] & 0x1f;
     spnt->rmvbl = (pagestart[1] & 0x80) >> 7;
+    spnt->wwid = no_wwid; spnt->serial = no_serial;
+    ansi = pagestart[2] & 7;
+	
+    if (ansi >= 3)
+	lun = 0;
+    else
+	lun = spnt->lun;
 
+    /* TODO: Extract serial number from bytes 36--43 ? */
+    
+    // List of supported EVPD pages ...
+    if (get_inq_page (infile, lun, buffer, 0, 1))
+	return 0;
+    ln = pagestart[3];
+    for (off = 0; off < ln; ++off) {
+	if (pagestart[4+off] == 0x80)
+		have_ser_page = 1;    
+	if (pagestart[4+off] == 0x83)
+		have_wwid_page = 1;    
+    }
+    
+    if (have_ser_page && !get_inq_page (infile, lun, buffer, 0x80, 1)) {
+	spnt->serial = getstr (pagestart, 4, 3+pagestart[3]);
+	if (verbose == 2) 
+	    printf ("Serial for %s: %s\n", spnt->name, spnt->serial);
+    }
+
+    if (have_wwid_page && !get_inq_page (infile, lun, buffer, 0x83, 1)) {
+	spnt->wwid = extract_wwid (pagestart);
+
+	if (verbose == 2)
+	    printf ("WWID for %s: %Lx\n", spnt->name, spnt->wwid);
+    }
+
+    //close(infile);
     return 0;
 #endif
 }
